/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  |           =====>>>>> 2002 <<<<=====             |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (input.c)
  
  Written by: Damian J. J. Farnell

  Date:       Wed Jul 17 12:03:05 BST 2002

  Purpose:    Inputs the CCM script data

  Changes:
      - cout replaced by printf, little improvements (Joerg Sep2005)

  ===================================================

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"

#if (PARALLEL_FLAG>0)
#include<mpi.h>
#endif

FILE *script;
FILE *escript;

// count read in lines to give precise error message
static int ccm_read_line=1;  // normally ok, not correct in all cases

// function prototypes.

int  moveAlongHome(FILE *fp, const char *searchstring); // obsolete 2010-05
char * nextAssign(FILE *fp); // replaces moveAlongHome JS-2010-05 
//  + strcmp(varname, "XXX")==0;
double  readInFloat(FILE *fp);
int   readInInteger(FILE *fp);

void   definePointsInBT(void);
void   subsetPrint(void);
void   readOperator(struct hamiltonian_defn op[13+1], FILE *f1,
                    const char *opname);
extern int hash_bits; // see collect.c

// JS1002: add args
void input(int run_option, char *ccmfile, int argc, char *argv[])
{
  
  int i, j ,k, bonds, expect, max_s2;
  int  max_number_cells_BT=0;
  
  // initialized with "" to avoid warnings from valgrind
  //     "jump depend from uninitialized variable"
  char input_filename[100]="";  // $(arg1) = ./workdir/name.ccm
  // char tempPath[100]="";    // old options, obsolete
  // char temp[100]="";
  char *varname=NULL; // pointer to next assinment variable
  char expectations_input_filename[100]=""; // $(arg1)_exp
  char configurations2[100]="";

  double dummyFloat;

  Control.file=1;    // tells program to save CCM equations to file.
  Control.output=0;  // default (scan .ccm file) JS1002
  Control.startx=0;  // default start vector (see solve.c)

  if (myid==0) {
    printf(" name=%s runoption=%d\n", ccmfile, run_option);
    printf("\n v0806: other files are build by adding a suffix\n\n");
    // especially for jobsystems
  }
  sprintf(input_filename,              "%s",    ccmfile);
  sprintf(expectations_input_filename, "%s_exp", ccmfile);

  script=fopen(input_filename,"r");
  if (!script) {
    if (!myid) {
      printf("input: Cannot Open %20s ...\n", input_filename);
      if (strchr(ccmfile,'/')==NULL)
        printf("v0806: Please include the path name to the filename\n");
    }
    ERR(1);
  }

  // read in `control' parameters first:

  if (!myid)
    printf("\nControl Parameters:\n");

  // first, scan from ccm-files (options may overwrite it later)
  // break loop after "solution strategy=" on first "number=" (unitcell)
  varname=nextAssign(script);
  if (strcmp(varname,"output level")==0) {
    Control.output=readInInteger(script);
  }
  
  for (i=1;i<argc-2;i++) {  // Jan2010-JS
    if (strstr(argv[i],"--output=")) Control.output=atoi(argv[i]+9);
    if (strstr(argv[i],"--startx=")) Control.startx=atoi(argv[i]+9);
  } 
  if (!myid) // JS1002: this if() should not depend from Control.output
    printf("Output_level= %d (verbose)\n", Control.output);
    //  0 = default
    // -1 = no files (make check + parallel jobs)

  for (i=1;i<argc-2;i++) {  // Feb2010-JS
    if (strstr(argv[i],"--hash_bits=")) hash_bits=atoi(argv[i]+12);
    if (hash_bits<2 || hash_bits>=8*(int)sizeof(long)) {
      if (!myid) printf("WARNING: hash_bits out of range, set to 20\n");
      hash_bits=20;
    }
    if  (hash_bits>=8*(int)sizeof(int) && !myid)
      printf("WARNING: hash_bits >= int_bits (untested code)\n"); 
  } 
  if (!myid) printf("hash_bits= %d\n", hash_bits); //  20 = default
  if (Control.output > 0 && myid ==0) {
    printf("Find_Excitations= %d     (Runoption=4x, 0=no, 1=yes)\n",
           Control.excitations);
  }

  // new by Jun08: (ToDo: remove and use option!)
  Control.greatExpectations=0;
#if 0 // may be an disadvantage for make check
  varname=nextAssign(script);
  if (0==strcmp(varname,"find generalised expectation values")) {
    Control.greatExpectations=readInInteger(script);
    printf("please remove obsolete 'find generalised expectation values'"
           " from .ccm-file and use option --expects=0..n\n");
  } else { ERR("unexpected assignment in ccm-file"); }
#endif 
  
  for (i=1;i<argc-2;i++) {  // Jan2010-JS
    if (strstr(argv[i],"--expects=")) Control.greatExpectations=atoi(argv[i]+10);
  } 
  // Joerg Jun08: better get it as an option than changing ccm file format
  // moveAlongHome(); Control.excitations=readInInteger();
  if (Control.output > 0 && myid ==0) {
    printf("Find Generalised Expectations = %d  (0=no, 1..n=yes)\n",
                Control.greatExpectations);
    if (Control.greatExpectations > 0)
      printf("expectations name   = %s\n", expectations_input_filename);
  }
  if (Control.greatExpectations > 0) {
    escript = fopen(expectations_input_filename,"r");
    if (!escript) {  // if that happens on some tasks we are in trouble
      if (!myid)
        printf("ERROR: open %s, skipping\n", expectations_input_filename);
      Control.greatExpectations = 0;
    }
  }

  //if (!ierr)  // skip if the "expectation" options was missed (old versions)
  //moveAlongHome(script,"work directory"); // fscanf(script,"%80s",tempPath);  // work directory= ...
  //moveAlongHome(script,"output file");    // fscanf(script,"%80s",temp);      // output file= ...
  // this options will be ignored, tempPath is unused (Nov09 obsolete)
  
  // sprintf(Control.name,"%s%s_id%03d", tempPath, temp, myid);
  // sprintf(Control.name, "%s_id%03d", ccmfile, myid); // "name.ccm_id$(id)"
  // JS0902: only use _MYID for splitted files (.terms_MYID)
  sprintf(Control.name, "%s", ccmfile); // JS0902: replace "name.ccm_id$(id)"

  if (myid==0)
    printf("Simulation Filename = %s\n"
           " v0806: path and file name options from input file are ignored\n",
            Control.name);

  // read in parameters for solve.c
  
  // now read in the `solve' variables

  if (increment!=0.0 && !myid)
    printf("WARNING: minimum/maximum/increment overridden by options\n\n");
  moveAlongHome(script,"minimum");   minimum=  readInFloat(script);
  moveAlongHome(script,"maximum");   maximum=  readInFloat(script);
  moveAlongHome(script,"increment"); increment=readInFloat(script);
  for (i=1;i<argc-2;i++) {  // Jan2010-JS
    if (strstr(argv[i],"--increment=")) increment=atof(argv[i]+12);
    if (strstr(argv[i],"--minimum="))   minimum=atof(argv[i]+10);
    if (strstr(argv[i],"--maximum="))   maximum=atof(argv[i]+10);
  } 
  if (Control.output > 0 && !myid) {
    printf("Minimum delta=  %20.10f\n", minimum);
    printf("Maximum delta=  %20.10f\n", maximum);
    printf("Increment delta=%20.10f\n", increment); // .15 is max with double
    if (increment != (double)((float)increment))
      printf("WARNING: increment has limited accuracy (float != double)\n");
  }

  // next two variables are minimum and maximum y values
  // used in plot.tcl but not in ccm C program.
  moveAlongHome(script,"minimumy"); dummyFloat=readInFloat(script);
  moveAlongHome(script,"maximumy"); dummyFloat=readInFloat(script);

  // dimension (is defined by the lattice, recomputed later, obsolete!)
  moveAlongHome(script,"dimension"); Params.dim=readInInteger(script);
  if (Control.output > 0 && !myid)
    printf("Dimension= %d\n", Params.dim);

  if (Params.dim != Dimension) {
    if (!myid)
    printf("Error: Constant.h Dimension need to be set to %d for this model\n"
           "  higher values are not allowed, we waste energy + time else\n",
           Params.dim);
    if (Params.dim > Dimension) ERR("Dimension problem");
  }

  // maximum spin number (defined by spins of uc)
  moveAlongHome(script,"spin"); dummyFloat = readInFloat(script);
  // Joerg: We store the twofold of Spin quantum number to have an integer
  //        this is faster and makes the program more compact
  // ToDo: its the maximum of spins in the unit cell and can be
  //       automatically calculated (see below), so its obsolete to set here
  Params.s2 = (int) (2*dummyFloat + 0.5);
  if (Control.output > 0 && myid ==0)
    printf("Maximum spin quantum number=%f\n",Params.s2 / 2.0);

  // Bounding Triangle

  // dummyInt=0;
  // ToDo: moveAlongHomeAny searching next "=" and decide if its
  //   the right string or some strings scipped.
  //    check last string was ... one of optional[obsolete], necessary
  //    or unknown (ignored)
  //  ToDo: bounding is obsolete!

  // JS-2010-05-31 removed
  // moveAlongHome(script,"bounding"); Params.btLength=readInInteger(script);
  // if (Control.output > 0 && myid ==0)
  //    printf("BT_Length=%d (bounding triangle)\n",Params.btLength);
  // JS: length of triangle block >= m ? 
  
  // SUBn-m: n= spin flips
  // SUBn-m: m= max. geom. distance
  moveAlongHome(script,"n"); Params.n=readInInteger(script);
  moveAlongHome(script,"m"); Params.m=readInInteger(script);
  if (argc >= 3) {  // allow additional options
    int i;
    for (i=1;i<argc-2;i++) {  // Jan2010-JS
      if (strstr(argv[i],"--subm=")) {
        Params.m=atoi(argv[i]+7);
        if (!myid) printf("Override CCM.Params.m by option: %s\n", argv[i]+7);
      }
      if (strstr(argv[i],"--subn=")) {
        Params.n=atoi(argv[i]+7);
        if (!myid) printf("Override CCM.Params.n by option: %s\n", argv[i]+7);
      }
    } 
  }
  if (Control.output > -1 && myid ==0) // essential debug info
    printf("SUB-n-m:  n= %d  m= %d  (MAX_N,M= %d %d)\n",
           Params.n, Params.m, MAX_SUBN, MAX_SUBM);

  if (Params.m > MAX_SUBM) ERR("Params.m > MAX_SUBM");
  if (Params.n > MAX_SUBN) ERR("Params.n > MAX_SUBN");
  
  
  moveAlongHome(script,"bonds"); Params.bonds=readInInteger(script);
  if (Control.output > 0 && myid ==0)
    printf("Bonds= %d\n", Params.bonds);

  moveAlongHome(script,"restriction"); restrictionClusters=readInInteger(script);
  for (i=1;i<argc-2;i++) {  // Jan2010-JS
    if (strstr(argv[i],"--restriction=")) restrictionClusters=atoi(argv[i]+14);
  }
  if (Control.output > 0 && myid ==0)
    printf("Restriction on Clusters: Type=%d\n",restrictionClusters);
  
  moveAlongHome(script,"solution strategy"); solutionStrategy=readInInteger(script);
  if (argc >= 3) {  // allow additional options
    int i;
    for (i=1;i<argc-2;i++) {  // Jan2010-JS
      if (strstr(argv[i],"--mom=")) {
        extern double momentum; // solve.c
        momentum=atof(argv[i]+6);
        if(!myid) printf("set momentum=%s (option)\n", argv[i]+6);
      }
      if (strstr(argv[i],"--method=")) {
        solutionStrategy=atoi(argv[i]+9);
        if(!myid) printf("set method=%s (option)\n", argv[i]+9);
      }
    } 
  }
  if (PARALLEL_FLAG>0 && solutionStrategy == 1 && npe != 1) {
    if (myid==0)
      printf("WARNING: solutionStrategy 1 (Newton-Raphson) dont work parallel\n"
            "switching to solutionStrategy = 0 (Direct Iteration)\n");
    // old code exit here, but we make life easier (JS Apr08)
    solutionStrategy=0;
  }
  if (Control.output > 0 && myid ==0) {
    if (solutionStrategy == 1)
      printf("Solve CCM equations using Newton-Raphson method (1)\n");
    else
      printf("Solve CCM equations using Direct Iteration method (0)\n");
  }

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Unit Cell:\n");


  for(i=0;i<=MAX_UNIT_CELL;i++) {  // i=0 is not used (JS0911 ToDo)
    UnitCell.x[i]=0.0;
    UnitCell.y[i]=0.0;
    UnitCell.z[i]=0.0;
    UnitCell.s2[i]=0; // 2*s
  }

  moveAlongHome(script,"number"); i=fscanf(script,"%d",&UnitCell.number);
  if (Control.output > 0 && myid ==0) {
    printf("Number of sites in Unit Cell = %d\n", UnitCell.number);
    printf("Positions & Spin Quantum Number:\n");
  }
  if (i!=1) { printf("Error: reading UnitCell.number\n"); exit(1); }

  // read in all sites of the unit cell
  if (UnitCell.number<1 || UnitCell.number>MAX_UNIT_CELL) ERR("UC.number");
  for(max_s2=1,i=1;i<=UnitCell.number;i++) {
    UnitCell.x[i]=readInFloat(script);
    UnitCell.y[i]=readInFloat(script);
    UnitCell.z[i]=readInFloat(script);
    UnitCell.s2[i]=(int)(2*fabs(readInFloat(script))+0.5);
    if (argc >= 3) {  // allow additional options
      int j;
      for (j=1;j<argc-2;j++) {  // Jan2010-JS
        if (strstr(argv[j],"--spin=")) {
          UnitCell.s2[i]=(int)(2*fabs(atof(argv[j]+7))+0.5);
          if (!myid) printf("Override spin[%d] by option: %s\n", i, argv[j]+7);
        }
      } 
    }
    //    moveAlongHome(script);
    if (max_s2 < UnitCell.s2[i])
        max_s2 = UnitCell.s2[i];
  }
  // normaly we have one line per site, but readInFloat does not detect line feeds
  ccm_read_line+=UnitCell.number;

  // check that spin length doesnt exceed maxspinlength (Joerg 2005)
  if (max_s2 != Params.s2) {
     if (!myid) {
       printf("\nWARNING input: (max_spin=%3.1f) !=  (spin=%3.1f),"
              "\n set Params.s2 automatically to max_spin\n",
              0.5*max_s2, 0.5*Params.s2);
       printf("--- Do you forgot to set spin length in the unit cell? ---\n");
     }
     Params.s2 = max_s2;
     // exit(1); // We dont need to worry, Joerg
  }

  // do not wast memory/CPUcache, JS1002 but triggers a older problem
  // JS1002: the concept of coding as sites within bounding triangle (BT) only
  //   works for some (regular?) lattices like square and triangular lattice
  //   it failes on j1j2_square_with_uc2 (try sym=1), where we have
  //   a bounding square (of m/uc+uc-1 length) on the bravais lattice
  //   (which is still a BT on the bound lattice)
  //   code may be simpler when we use always the square,
  //   (wasting up to 1/2 at 2D and 5/6 at 3D for memory) for config descriptors
  //  may be direct coding is better for 3D m=20 20^3/6 = 1200 (no sub-2-100)
  //     minbits(to normalize before storing?)
  //     1D-sub-20-20: cfg=20*5bit=100bit     descr= 20bit
  //  s1-1D-sub-20-10: cfg=20*4bit=80bit      descr= 20bit
  //     2D-sub-14-14: cfg=14*(2*4)bit=48bit  descr= 14*13/2=91bits
  //     3D-sub-2-100: cfg=2*3*7bit=42bit     descr= 100*100*100/6bit=170000bits
  // temporary we set to worst case btLength=2(m/uc+uc-1)+1
  // probably a good idea to keep 
  // Params.btLength >= Params.m/UnitCell.number+1;
  if (UnitCell.number < 2 && Params.btLength < Params.m) {
    Params.btLength = Params.m; // Joerg Oct05, set bounding triangle to m
    if (!myid)
     printf("\nWARNING: set min bounding=btLength= %d\n", Params.btLength);
  }
  // this does not work in every case!!! for example if 2nd site is a poor
  //   attachement of square lattice sites, but usually it should work
  if (UnitCell.number > 1
   && Params.btLength < 2*Params.m/UnitCell.number+UnitCell.number-1) {
      Params.btLength = 2*Params.m/UnitCell.number+UnitCell.number-1;
    if (!myid)
      printf("\nWARNING: set min bounding=btLength= %d\n", Params.btLength);
  }
  if ((UnitCell.number > 1 && Params.btLength < 2*Params.m/UnitCell.number
                                                          +UnitCell.number-1)
    ||(UnitCell.number < 2 && Params.btLength > Params.m)) { // JS2010-03
    if (!myid)
     printf("\nWARNING: bounding is to big (wasting memory and CPU time)\n"
       " ... please use auto setting (bounding= 0)\n");
  }
  // BT = bounding triangle  
  // number of different elements of a symmetric multidimensional matrix
  // JS1002: is this also correct (sufficiant) for triangular lasttice?
  //         or square without symmetries? or J2-lattice?
  //    .        .        7           4
  //    . .      . .      . 5        . 3
  //    . 3 4    3 3 .    . . 5     . . 3
  //    4 3 . .  3 3 . .  . . . 7  . . . 4  m=4
  //    sq-4-4   sq-4-3   sq-4-7   tr-4-4
  //  (show max dist of sites to other sites)

  max_number_cells_BT = 1;  // js0903 used for animals?
  { int i;
    for (i=0; i<Params.dim; i++) max_number_cells_BT *= (Params.btLength+i);
    for (i=0; i<Params.dim; i++) max_number_cells_BT /= (i+1);
  }
  if (Control.output > -1 && myid ==0) 
    printf("\nDimension=%d  2s=%d  maxdist=subm=%d\n"
           "  btLength=sizeLUT=%d  numSites_in_BT=%d (bounding triangle)\n\n",
	   Params.dim, Params.s2, Params.m, Params.btLength,
	   max_number_cells_BT);
  // JS1002: max_number_cells_BT was/is not used

  initDescr(Params.s2);  // store 0! .. (2s+1)! into array

  for(i=1;i<=UnitCell.number;i++) {
    if (Control.output > 0 && myid ==0)
      printf("Site= %2d : position = (% f,% f,% f) : spin = %f\n",i,
        UnitCell.x[i], UnitCell.y[i], UnitCell.z[i], 0.5*UnitCell.s2[i]);
  }

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Bravais Lattice:\n");

  // define the bravais lattice 

  moveAlongHome(script,"");  // marked by "=" only followed by 3 numbers and 2 lines
  for (i=0; i<3; i++) // assume 3D
  for (j=0; j<3; j++)
    BravaisLattice[i][j]=readInFloat(script); // old0903: .{a,b,c}[1..3]
  // normaly we have one line per site, but readInFloat does not detect line feeds
  ccm_read_line+=3-1;
  
  if (Control.output > -1 && myid ==0) {
    for (i=0; i<3; i++)
      printf("Bravais Lattice Vector %c=(% 6.4f,% 6.4f,% 6.4f)\n", 'A'+i,
  	BravaisLattice[i][0], BravaisLattice[i][1], BravaisLattice[i][2]);
  }
  { // JS0903: compute inverse matrix to solve A.x = b later
    //  invers = adj(A)/det(A),  adj(A)=transpose(subDet)
    double det, non_ortho=0.0, (*m)[3]=BravaisLattice;
    for (i=0; i<3; i++)
      non_ortho += fabs(m[0][i]*m[1][i])
                 + fabs(m[1][i]*m[2][i])
                 + fabs(m[2][i]*m[0][i]);
    det =  m[0][0]*m[1][1]*m[2][2]
         + m[0][1]*m[1][2]*m[2][0]
         + m[0][2]*m[1][0]*m[2][1]
         - m[0][0]*m[1][2]*m[2][1]
         - m[0][2]*m[1][1]*m[2][0]
         - m[0][1]*m[1][0]*m[2][2];
    if (Control. output > -1 && myid ==0) 
      printf(" Bravais Lattice Orthogonality if 0 == %g\n"
             " Det([A,B,C])= %g\n", non_ortho, det);
     inversBravaisLattice[0][0] = (m[1][1]*m[2][2]-m[1][2]*m[2][1])/det;
     inversBravaisLattice[0][1] = (m[0][2]*m[2][1]-m[0][1]*m[2][2])/det;
     inversBravaisLattice[0][2] = (m[0][1]*m[1][2]-m[0][2]*m[1][1])/det;
     inversBravaisLattice[1][0] = (m[1][2]*m[2][0]-m[1][0]*m[2][2])/det;
     inversBravaisLattice[1][1] = (m[0][0]*m[2][2]-m[0][2]*m[2][0])/det;
     inversBravaisLattice[1][2] = (m[0][2]*m[1][0]-m[0][0]*m[1][2])/det;
     inversBravaisLattice[2][0] = (m[1][0]*m[2][1]-m[1][1]*m[2][0])/det;
     inversBravaisLattice[2][1] = (m[0][1]*m[2][0]-m[0][0]*m[2][1])/det;
     inversBravaisLattice[2][2] = (m[0][0]*m[1][1]-m[0][1]*m[1][0])/det;
  }
  if (Control.output > -1 && myid ==0) {
    for (i=0;i<3;i++)
      printf("invers Bravais Lattice Vector i%c=(% 6.4f,% 6.4f,% 6.4f)\n",
              'A'+i,
  	      inversBravaisLattice[i][0],
              inversBravaisLattice[i][1],
              inversBravaisLattice[i][2]);
  }

  if (Control.output > 0 && myid ==0) {
    printf("=======================================\n"
	   "Nearest-Neighbours to sites in Unit Cell\n");
    printf(" (define distance m and LatticeNeighbours to compute Lattice animals from it):\n");
  }
  // needed to decouple lattice configurations from parameters (used in define.c)

  if (UnitCell.number<1
   || UnitCell.number>MAX_UNIT_CELL) ERR("UC.number");
  for(i=1;i<=UnitCell.number;i++) {
    moveAlongHome(script,"number");
    LatticeNeighbours.number[i]=readInInteger(script);
    if (LatticeNeighbours.number[i]<0
     || LatticeNeighbours.number[i]>=MAX_NUM_NEIGHBOURS) ERR("maxNN");
    for(j=1;j<=LatticeNeighbours.number[i];j++) {
	  LatticeNeighbours.x[i][j]=readInInteger(script);
	  LatticeNeighbours.y[i][j]=readInInteger(script);
	  LatticeNeighbours.z[i][j]=readInInteger(script);
	  LatticeNeighbours.cell[i][j]=readInInteger(script);
    }
  }
  // normaly we have one line per site, but readInInteger does not detect line feeds
  ccm_read_line+=UnitCell.number+1;

  if (Control.output > 0 && myid ==0) {
    for(i=1;i<=UnitCell.number;i++) {
      printf("Number of nearest-neighbours for site %d = %d : \n",
  	      i, LatticeNeighbours.number[i]);
      for(j=1;j<=LatticeNeighbours.number[i];j++) {
        // ToDo: Neighbour = Bravais translation? wrong for ferriLC (JS06) 
	printf(" %d : Bravais translation (x y z site) = (% d % d % d % d)\n",
	       j, LatticeNeighbours.x[i][j],
                  LatticeNeighbours.y[i][j],
	          LatticeNeighbours.z[i][j],
                  LatticeNeighbours.cell[i][j]);
      }
      for(j=1;j<=LatticeNeighbours.number[i];j++) {
        // JS08: lc_j1j2j3 dim=1 but nearest-neighbours of uc in y direction
        //       cause error of recursiveClusteringFunction()
        if ((LatticeNeighbours.z[i][j] && Params.dim<3)
         || (LatticeNeighbours.y[i][j] && Params.dim<2)) {
          printf(" Error: dimension and lattice neighbours contradict\n");
          MPI_Abort(MPI_COMM_WORLD, 1);
        }
      }
    }
    printf("=======================================\n"
	   "The symmetry operations:\n");
  }
  moveAlongHome(script,"number");
  LatticeSymmetries.number=readInInteger(script); 

  if (LatticeSymmetries.number<1) {
    printf("input: we need at least the one symmetry operator (identity)\n"
           "  to get findDescriptorOfCluster running\n"); exit(1);
  }
  if (LatticeSymmetries.number>MAX_NUM_SYM) {  // JS1003
    printf("input: LatticeSymmetries.number>=MAX_NUM_SYM\n"); exit(1);
  }
  // just to make sure set to zero initially.

  // JS-2010-03-22: BUG fixed, MAX_NUM_SYM==LatticeSymmetries.number
  for(i=0;i<LatticeSymmetries.number;i++){  // JS1003: 0..n-1 old=1..n
    // JS Jun08+1003: change .x,.y,.z to LS.op[i].R[0..2][3]
    for(j=0;j<3;j++)
      for(k=0;k<4;k++) // ROT+TRANS
	LatticeSymmetries.R[i][j][k]=0.0;

    for(j=0;j<3;j++)
      for(k=0;k<4;k++)
	LatticeSymmetries.R[i][j][k]=readInFloat(script);
  // normaly we have one line per site, but readInFloat does not detect line feeds
  ccm_read_line+=3;

    if (Control.output > 0 && myid ==0) {
      printf("Symmetry Operation = %d\n"
             "Rotation matrix : Prim. Transl. \n",i);
      for(j=0;j<3;j++){
	for(k=0;k<3;k++)
	  printf( "% 5.3f ", LatticeSymmetries.R[i][j][k]);
	printf(" : % 5.3f\n",LatticeSymmetries.R[i][j][3]);
      }

      // ToDo: insert bravais-sym-computation here, BravaisSymmetries.B[][][]
      //  rotation in bravais coordinates + permutation of sites in unitcell
      //     for each site in unit cell per symmetry
      // (ROT,TRANS,UC) for each Sym*UC
      //  kagome?  mc_lf

    } 
  }
  if (argc >= 3) {  // allow additional options
    int j, nsym;
    for (j=1;j<argc-2;j++) {  // Jan2010-JS
      if (strstr(argv[j],"--symmetries=")) {
        nsym=atoi(argv[j]+13);
        if (nsym<=LatticeSymmetries.number && nsym>=1) {
          LatticeSymmetries.number=nsym;
          if (!myid) printf("Override num symmetries: %s\n", argv[j]+13);
        } else
          if (!myid) printf("ERROR set symmetries=%s failed\n", argv[j]+13);
      }
    } 
  }

  // now determine the Hamiltonian. The `types' of term
  // first...
  
  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Hamiltonian:\n");
  
  moveAlongHome(script,"directional");
  directional=readInInteger(script);
  //  1 for kagome_j1_j2, kagome_triangle, lc_mf, maple_bounce*,
  //        square_external_field, scrubo_spiral, triangle_external_field*
  if (!myid)
    printf("Directional?=%d  (1=directional, 0=non-directional bonds)\n",
           directional);

  if (Control.output>0 && !myid) {
    printf("output operator terms (abbrv: zz=s^z.s^z z+=s^z.s^+ etc.):\n");
  }
  // JS1002: dynamic memory allocation
  Hamiltonian = (struct hamiltonian_defn (*)[13+1])
    malloc((Params.bonds+1)*(13+1)*sizeof(struct hamiltonian_defn));
  if (!Hamiltonian) ERR("malloc Hamiltonian");
  //  JS1002: using general readOperator-function
  for (bonds=1;bonds<=Params.bonds;bonds++) {
    if (!myid) printf("read_Bond %2d\n", bonds);
    readOperator( Hamiltonian[bonds], script, "Bond");
  }

  // now input the information for the Hamiltonian proper for each term
  // in the similarity transformed version of H.
  // Jun08 Joerg: transformation removed for simplification

  fclose(script);

  // 5-5-08: Additional code to implement generalised expectation values 
  if (Control.greatExpectations > 0) {
    moveAlongHome(NULL,""); /* reset counter */
 
    // first: the number of expectation values
    moveAlongHome(escript,"number of expectations"); Params.nExpect=readInInteger(escript);
    if (Control.output > 0 && myid ==0)
      printf("\n\nNumber of Expectation Values = %d \n",Params.nExpect);
    if (Params.nExpect > Control.greatExpectations)
        Params.nExpect = Control.greatExpectations; // JS-2010-05-30
    // this will generate incompatibility before 2010-05-30
    //   only one expectation value is computed on old versions 
    // JS1002: dynamic memory allocation
    expectation = (struct hamiltonian_defn (*)[13+1])
      malloc((Params.nExpect+1)*(13+1)*sizeof(struct hamiltonian_defn));
    if (!expectation) ERR("malloc expectation");
    //  JS1002: using general readOperator-function
    for (expect=1;expect<=Params.nExpect;expect++) {
      if (!myid) printf("read_Expect %2d\n", expect);
      readOperator( expectation[expect], escript, "Expect");
    }
  }  // ControlGreatExpectations
  
  definePointsInBT();

  //  init x_array[],x_index[],cell_index[]
  //  defineGeneralisedPositions();  -> removed (not used, Joerg05)

#if 0
  // not used at moment...
  sublatticeMagnetisation.number=UnitCell.number;
  if (sublatticeMagnetisation.number<0
   || sublatticeMagnetisation.number>MAX_UNIT_CELL) ERR("subL");
  for (i=1;i<=sublatticeMagnetisation.number;i++)
    sublatticeMagnetisation.cell[i]=1;
#endif
 
  // read in cluster configurations 
  
  if ((run_option&1) == 0) { // loading configs from file, else compute
    int nread;

    sprintf(configurations2, "%s.configs", Control.name);
    if (!myid) 
      printf("\nReading ConfigDescriptors(n*%d*(int=%d)) from file %s\n",
                      DESCR_SIZE, (int)sizeof(int), configurations2);
    if (!myid) {  // JS Apr08: read only once and distribute per MPI
      fp1=fopen(configurations2,"r"); // $(Control.name).configs  
      if (!fp1) {
        printf("input: Cannot Open %s... Exiting!!\n",configurations2);
        exit(1);
      }
      // pre060611: fread(&NumberConfigurations,sizeof(int),1,fp1);    
      for(i=1;i<=MAX_NUM_CONFIGURATIONS;i++) {

        nread=fread(spinConfigurations[i].descriptor,
               ((DESCR_SIZE)*sizeof(int)), 1, fp1);
        if (nread!=1) {
           if (feof(fp1)) { i--; break; }
           printf("%d: ferror %d\nExit!\n",myid,ferror(fp1));
           exit(1);
        }
        if (Control.output > 2 && myid==0 && i<10){
          printf("Cluster %d:\n",i);
          printOutClusterFromDescriptor(spinConfigurations[i].descriptor);
        }
        if (feof(fp1)) break;
      }
      fclose(fp1);
    }
    NumberConfigurations = i;

    if (!myid) {
      printf("\nFundamental clusters = %d\n", NumberConfigurations);
    }
#if PARALLEL_FLAG>0
    int iret=0;
    if (Control.output>0 && !myid)
      printf(" calling MPI_Bcast(.configs, (int)(%d*%d)) ...\n",
             (int)sizeof(struct spinConfigs), i); 
    iret= MPI_Bcast(&NumberConfigurations, 1, MPI_INT, 0, MPI_COMM_WORLD);
    if (iret!=MPI_SUCCESS) MPI_Abort(MPI_COMM_WORLD, 1);
    i = NumberConfigurations;
    iret= MPI_Bcast(spinConfigurations+1,
               sizeof(struct spinConfigs)*i, MPI_BYTE, 0, MPI_COMM_WORLD);
    if (iret!=MPI_SUCCESS) MPI_Abort(MPI_COMM_WORLD, 1);
#endif

  }
  if (!myid) subsetPrint(); // output subset rule
}


// Joerg Jun08: return pointer to varname, if found, NULL if failed
char * nextAssign(FILE *fp)
{
  char c1=0;
  // must be static, a pointer will be returned
  static char buff[2000]=""; // tcc-0.9.25 -b needs =" " to go around a bug
  int  count, bp=0, iscomment=0, nread;
  /* 
     renamed old moveAlongHome() (2010-05)
     purpose of this routine is to scan along the file until
     an `=' sign is seen 
     Joerg Sep05: ignore out commented lines (// ... and # ...)
     Joerg Jun08: add searchstring, for better debugging (new control parameter)
     Joerg Mai10: return varname and check on caller
  */

  /* reset */  
  if (!fp) { ccm_read_line=1; return NULL; }

  buff[0]=0;
  for(count=0;count>-1;count++) {
    if (feof(fp)) { count=-1; break; }
    nread=fscanf(fp,"%c",&c1);              // char by char
    if (nread!=1) { count=-1; break; }
    if (c1 != '\n' && (c1 != '=' || iscomment)) {
      buff[bp]=c1; bp++; buff[bp]=0; // add a char to the buffer
      if (bp>=2000) { count=-2; break; }
    }
    // fprintf(stderr," read: %c buff=%s bp=%d\n",c1,buff,bp);
    // printf("L708 Control.output=%p\n", &Control.output);
    if (Control.output > 2 && myid == 0) {
      if (c1=='\n' && buff[0])    printf("%% %3d: %s\n",      ccm_read_line, buff);
      if (c1=='='  && !iscomment) printf("%% %3d: %s= ...\n", ccm_read_line, buff);
    }
    if (bp>1 && buff[bp-1]=='/'
             && buff[bp-2]=='/') iscomment=1;
    if (c1=='#')                 iscomment=1;
    if (c1=='\n') {
      if (myid==0 && (!iscomment) && ((int)strspn(buff," \t\n\r"))!=bp) // JS Oct09
        printf("WARNING line %3d, unexpected string: %s\n",ccm_read_line,buff);
      iscomment=0; ccm_read_line++; bp=0; buff[bp]=0;
    }
    if (c1 == '=' && !iscomment) { break; } // found next "XXX="
  }
  if (count < 0) { 
    ERR("reading CCM script: can't find delimiter! EOF reached.");
  }
  return buff;
}


// Joerg Jun08: return 0, if found, -1 if failed (obsolete)
//   better do loop and case construct JS10
int moveAlongHome(FILE *fp, const char *searchstring)
{
  char *buff;
  while (1) {
    buff=nextAssign(fp);
    if (buff==0 || (!searchstring[0]) || strcmp(buff,searchstring)==0)  // JS1002
      break;  // found, if searchstring=="" accept everything
    else {
      if (!myid)
        printf("ERROR: searched for \"%s=\" but found \"%s=\" (line=%d),"
               " skip it\n", searchstring, buff, ccm_read_line);
    }
  }
  return 0;
}



// ToDo: unfortunatly we cant detect EOL using fscanf to get ccm_read_line up to date
//   use global linebuffer and pointer to next word

int readInInteger(FILE *fp) {
  int sum=0; int num_input=0;
  static char ss[200]="", // tcc-0.9.25 -b needs ="" to go around a bug
              *rets=NULL;
  while(1) {  // Oct09 JS
    num_input=fscanf(fp,"%d",&sum);
//    printf("readInInt1: ret=%d val=%d\n", num_input, sum);
    if (num_input==1) break; // OK, proceed
    
    ss[0]=0;
    rets = fgets(ss, 200, fp); // read until EOL incl. \n
    if ((!rets) || (ss[0]!='#' && (ss[0]!='/' || ss[1]!='/'))) {
      printf("ERROR readInInteger line=%d, read: %s", ccm_read_line, ss);
      MPI_Abort(MPI_COMM_WORLD,1);
    }
  }
  return sum;
}
    
double readInFloat(FILE *fp) {
  double sum=0.0; int num_input=0;
  static char ss[200]="", // tcc-0.9.25 -b needs ="" to go around a bug
              *rets=NULL;
  while(1) { // Oct09 JS
    num_input=fscanf(fp,"%lf",&sum);
//    printf("readInFloat1: ret=%d val=%f\n", num_input, sum);
    if (num_input==1) break; // OK, proceed

    ss[0]=0;
    rets = fgets(ss, 200, fp); // read until EOL incl. \n
    if ((!rets) || (ss[0]!='#' && (ss[0]!='/' || ss[1]!='/'))) {
      printf("ERROR readInFloat line=%d, read: %s", ccm_read_line, ss);
      MPI_Abort(MPI_COMM_WORLD,1);
    }
  }
  return sum;
}


// see initDescr
extern int sites_per_int; // old: extern int block_size;


// called by input() = initBTarray (BT = bounding triangle)
// initialize: lookUpTable[0..sizeLUT-1]  (old MAX_BT)
//                 used by [printOut,find]ClusterFromDescriptor
//             BTarray[z][y][x] // memory: m^3*int
// replaced by BTarray[z][y]+x  // memory: m^2*int
// replaced by BTarray[z*m+y]+x // malloc(m^(dim-1)*int) (ToDo?)
//                 used by findConfigurationToDescriptor
//                 0 <= x,y,z < Params.btLength
// ToDo: estimate size and malloc lookUpTable, simplify (Joerg)
//   sample: btLength=m=4 D=2
//   (y,x)=(0,0),(0,1),(0,2),(0,3), = 0,1,2,3,
//         (1,0),(1,1),(1,2),         4,5,6,
//         (2,0),(2,1),               7,8,
//         (3,0)                      9                 
//                                  
//    n*(n+1)/2 = 4*5/2 = 10 entries
//    1D:  n/1  2D: n*(n+1)/2, 3D: n*(n+1)*(n+2)/6
// rewritten by Joerg Oct05         
void definePointsInBT(void)
{
  int i, x,y,z, count,mx,my,mz;
  
#if 0 // JS1002 not needed anymore!?
  if (Params.btLength > MAX_SUBM) {
    printf("Params.btLength=%d > MAX_SUBM\n", Params.btLength);
    MPI_Abort(MPI_COMM_WORLD,1);
  }
#endif
#if 0 // obsolete and bad! N=6 s=1 sub-6-12  JS1002
  if (Params.btLength < Params.m) {
    printf("Params.btLength=%d < Params.m\n", Params.btLength);
    MPI_Abort(MPI_COMM_WORLD,1);
  }
#endif

  mx=my=mz=Params.btLength; // maximum x,y,z for 3D-case
  if (Params.dim < 3) mz=1; // max_z=0 for 2D (periodic 2D lattice)
  if (Params.dim < 2) my=1; // max_y=0 for 1D (periodic 1D lattice)
  if (Params.dim < 1) mx=1; // max_x=0 for 0D (finite cluster = UC)

  count=0; // site index, starting from 0 (Oct05)
  for (z=0;z<mz    ;z++) 
  for (y=0;y<my-z  ;y++) 
  for (x=0;x<mx-y-z;x++) { // (x,y,z) <=> idx, replacable by 2 functions ???
    count++;               
  }
  //  1D: n/1,  2D: n*(n+1)/(1*2),  3D: n*(n+1)*(n+2)/(1*2*3)
  { int count2=1;
    for (i=0;i<Params.dim;i++) count2 *= (Params.btLength+i);
    for (i=0;i<Params.dim;i++) count2 /= (i+1);
    if (count != count2) {   // can not happen! but we check for safety
      printf("ERROR definePointsInBT count!=count2\n");
      MPI_Abort(MPI_COMM_WORLD,1);
    }
  }
  sizeLUT = count;
  lookUpTable = (int (*)[3]) malloc(sizeLUT*sizeof(int)*3);
  if (!lookUpTable) {
      printf("ERROR definePointsInBT lookUpTable malloc(%d*int)\n", sizeLUT);
      MPI_Abort(MPI_COMM_WORLD,1);
  }

  BTarray = (int *) malloc( mz*my*sizeof(int) );
  if (!BTarray) {
      printf("ERROR definePointsInBT BTarray malloc(%d*int)\n", mz*my);
      MPI_Abort(MPI_COMM_WORLD,1);
  }

  count=0; // site index, starting from 0 (Oct05)
  for (z=0;z<mz    ;z++) 
  for (y=0;y<my-z  ;y++) 
  for (x=0;x<mx-y-z;x++) { // (x,y,z) <=> idx, replacable by 2 functions ???
    if (z*Params.btLength+y >= mz*my) ERR("sizeBT");
    if (x==0) BTarray[z*Params.btLength+y]=count;
    if (count >= sizeLUT) ERR("sizeLUT");
    lookUpTable[count][0]=x;
    lookUpTable[count][1]=y;
    lookUpTable[count][2]=z;
    count++;                    //  2D: n*(n+1)/2, 3D: n*(n+1)*(n+2)/6
  }
  // added by Joerg Sep2005  cubic 10 -> 10*12^2/4=1440/4=385=10*11/2*7
  // set this constants to reduce memory needs to minimum
  {
    int max_spin_list = (sizeLUT*UnitCell.number + sites_per_int - 1)/sites_per_int;
    if (Control.output > 0 && myid==0) { // JS1002: BSize to sites_per_int
      printf("\ndefinePointsInBT: btLength= %d sizeLUT= %d UC= %d sites_per_int= %d\n",
          Params.btLength, sizeLUT, UnitCell.number, sites_per_int);
      printf("\ndefinePointsInBT: DESCR_SIZE= %d (Constants.h)\n",
          max_spin_list);
    }

    if (max_spin_list > DESCR_SIZE) {
      printf("Error: Constant.h DESCR_SIZE < %d\n", max_spin_list);
      MPI_Abort(MPI_COMM_WORLD, 1);
    }
  }
}  


// read Hamilton and read expectations does the same, so we use one function
//  JS1002
// for (bonds=1;bonds<=Params.bonds;bonds++) {
//   if (!myid) printf("read_Bond= %2d\n", bonds);
//   readOperator( Hamiltonian[bonds], script, "Bond"); }
// for (expect=1;expect<=Params.nExpect;expect++) {
//   if (!myid) printf("read_Expect= %2d\n", expect);
//   readOperator( expectation[expect], escript, "Expect"); }
void readOperator(struct hamiltonian_defn op[13+1], FILE *f1,
  const char *opname) {
  const char *tdescr[13+1]={"none",  
    "zz", "z+", "z-", "+z", "-z", "+-", "-+", "++", "--", // 1..9
    "z", "-", "z^2", "+"};                // 10..13   Oct05 Joerg
  int i, j, k;
  // initially set everything zero
  for(k=1;k<13;k++) {  // Jun08 Joerg: 13 types of terms
    op[k].evaluate=0;
    for(i=0;i<=MAX_UNIT_CELL;i++) {
      op[k].k_to_m_number[i]=0;
      op[k].m_to_k_number[i]=0;
      for(j=0;j<=MAX_NUM_NEIGHBOURS;j++) {
        op[k].k_to_m[i][j][0]=0; // x
        op[k].k_to_m[i][j][1]=0; // y
        op[k].k_to_m[i][j][2]=0; // z
        op[k].k_to_m[i][j][3]=0; // cell
        op[k].m_to_k[i][j][0]=0; // x   
        op[k].m_to_k[i][j][1]=0; // y
        op[k].m_to_k[i][j][2]=0; // z
        op[k].m_to_k[i][j][3]=0; // cell
        op[k].coefficient[i][j]=0.0;
      }
    }
  }
  for(i=1;i<=UnitCell.number;i++) {
    for(k=1;k<=13;k++) {    /* 1..13 */

      moveAlongHome(f1,"number");
      op[k].k_to_m_number[i]=readInInteger(f1);
      if (op[k].k_to_m_number[i]>=MAX_NUM_NEIGHBOURS)
              ERR("numNeighbours>=MAX_NUM_NEIGHBOURS");

      if (k<=9) {  /* 1..9 are bonds */

        /* read in forward directions */
        for(j=1;j<=op[k].k_to_m_number[i];j++) {
          op[k].k_to_m[i][j][0]=readInInteger(f1);
          op[k].k_to_m[i][j][1]=readInInteger(f1);
          op[k].k_to_m[i][j][2]=readInInteger(f1);
          op[k].k_to_m[i][j][3]=readInInteger(f1);
          op[k].coefficient[i][j]=readInFloat(f1);
          if (op[k].k_to_m[i][j][3]<1
           || op[k].k_to_m[i][j][3]>UnitCell.number)
          printf("ERROR: site index out of range (1..nUC)\n");
          // normal input line: x y z uc coeff
          ccm_read_line+=1;
        }
        if (directional) { /* two directions, read in backwards direction */

          moveAlongHome(f1,"number"); 
          op[k].m_to_k_number[i]=readInInteger(f1);
          if (op[k].m_to_k_number[i]>=MAX_NUM_NEIGHBOURS)
                  ERR("numNeighbours>=MAX_NUM_NEIGHBOURS");

          for(j=1;j<=op[k].m_to_k_number[i];j++) {
            double rr;
            op[k].m_to_k[i][j][0]=readInInteger(f1); // x
            op[k].m_to_k[i][j][1]=readInInteger(f1); // y
            op[k].m_to_k[i][j][2]=readInInteger(f1); // z
            op[k].m_to_k[i][j][3]=readInInteger(f1); // c
            if (op[k].m_to_k[i][j][3]<1
             || op[k].m_to_k[i][j][3]>UnitCell.number)
            printf("ERROR: site index out of range (1..nUC)\n");
            rr=readInFloat(f1);
            ccm_read_line+=1;
            // JS0812 kagome_j1_j2.ccm: forward-backward<=1e-9
            if (fabs(rr-op[k].coefficient[i][j])>1e-8
                    &&  op[k].m_to_k_number[i]
                     <= op[k].k_to_m_number[i]
                    && !myid)
              printf("Warning: backward != forward Coeff"
                     " (type=%d,site=%d,%.9f != %.9f)\n", k, i,
                      op[k].coefficient[i][j], rr);
            op[k].coefficient[i][j]=rr;
            /* ToDo: overwiting same coefficient? */   
          }

        } else {
          /* one direction, copy forwards to backwards direction */

          op[k].m_to_k_number[i]= op[k].k_to_m_number[i];

          for(j=1;j<=op[k].k_to_m_number[i];j++) {
            op[k].m_to_k[i][j][0]=op[k].k_to_m[i][j][0];
            op[k].m_to_k[i][j][1]=op[k].k_to_m[i][j][1];
            op[k].m_to_k[i][j][2]=op[k].k_to_m[i][j][2];
            op[k].m_to_k[i][j][3]=op[k].k_to_m[i][j][3];
          }
        }  
      } else {  /* 10..13 are single site terms */
        if (op[k].k_to_m_number[i]) { // Joerg Aug09
            op[k].coefficient[i][1]=readInFloat(f1);
        if (op[k].k_to_m_number[i]!=1) {
            ERR("number for single site terms can only be 0 or 1");
          }
          ccm_read_line+=1;
        }
      }  
    }    
  }
  /* set flag to evaluate if we have nonzero coefficient (gt 0.01) */
  //        until 0605 only coefficients bigger than fabs(0.01) are used
  // Joerg Jun08 ToDo: what about m_to_k? a bug?
  for(k=1;k<=13;k++) {
    op[k].evaluate=0;
    for(i=1;i<=UnitCell.number;i++) {
      for(j=1;j<=op[k].k_to_m_number[i];j++) {
        if (op[k].coefficient[i][j] != 0.0) { op[k].evaluate=1; break; }
      }  
      if (op[k].evaluate) break;
    }
  }
       
  // now print out this information:
  if (Control.output > 0 && myid ==0) {
    for(k=1;k<=13;k++) {
      if (!op[k].evaluate) continue; // Oct05
      // space before next type of bonds
      // JS1002: remove " :" to get line below 80 columns
      if (k > 9) {  /* single site term */
        for(i=1;i<=UnitCell.number;i++) { 
          if (op[k].coefficient[i][1])
          printf(" Type= %2d Site= %2d Coeff= % .5f %s\n",
                 k, i, op[k].coefficient[i][1], tdescr[k]);
        }
      } else {  /* k<=9 bonds */
        for(i=1;i<=UnitCell.number;i++) {
          for(j=1;j<=op[k].k_to_m_number[i];j++) {
            printf(" Type= %2d Site=%2d "
                   "xyzc= % d % d % d % d Coeff= % .5f k-m %s\n", k, i,
                   op[k].k_to_m[i][j][0], op[k].k_to_m[i][j][1],
                   op[k].k_to_m[i][j][2], op[k].k_to_m[i][j][3],   
                   op[k].coefficient[i][j], tdescr[k]);
          }
          if (!directional) continue;
          for(j=1;j<=op[k].m_to_k_number[i];j++) {
            printf(" Type= %2d Site=%2d "
                   "xyzc= % d % d % d % d Coeff= % .5f m-k %s\n", k, i,
                 op[k].m_to_k[i][j][0], op[k].m_to_k[i][j][1],
                 op[k].m_to_k[i][j][2], op[k].m_to_k[i][j][3],
                 op[k].coefficient[i][j], tdescr[k]);
          } // m_to_k_number
        } // UC  
      } // if k=1..9 + 10..13
    } // k=1..13 
  }          
}
