/* 
                    |-------------------------|
                    | PROGRAM NAME: SrCuBO.C  |
                    |                         |
                    |  Copyright: R.Darradi   |
                    |  Modified: Joerg S. 2009|
-------------------------------------------------------------------------------
   Shastry-Sutherand model (SrCu2(BO3)2). Sha_Suth_model.ccm

   use ../Data/srcubo_spiral.ccm 
   set USR_FILE to sample_SrCuBO.c in solve.c

    *-------*-------*-------*-------*
    |      /|       |      /|       |
    |  J'/  |       |    /  |       |
  J |  /    |       |  /    |       |
    |/      |       |/      |       |
    *-------*-------*-------*-------*
    |   J   |\      |4     3|\      |
    |       |  \    |       |  \    |
    |       |    \  |       |    \  |
    |       |      \|1     2|      \|
    *-------*-------*-------*-------*
    |      /|       |      /|       |
    |    /  |       |    /  |       |
    |  /    |       |  /    |       |
    |/      |       |/      |       |
    *-------*-------*-------*-------*

    bond1 = J = Delta ??    JS09: bond1 = J' = Delta
    bond2 = J'              JS09: bond2 = J  = 1
*/

#define _USR_PARAMETER_ "bonds1_square=D bonds2_diag=1" // JS0910
double usr_parameter(int bonds, double alpha)
{
  /* add your model specifique parameter here! H=sum_i u(i,alpha)*H_i */
  //  d1 * H1 + d2 * H2 + d3 * H3 + ...
  if (bonds==1) return alpha; // field
  if (bonds==2) return 1;     // nn
  return 0.0;
}

#define UseAlpha 1
#define StartAlpha    (0.23408146*M_PI/2.0*2) // min(D=1.58)
#define StepAlpha     (0.01*M_PI)    // initial step of alpha, 0.001*M_PI
#define SearchAlphaMin 2 // 0=startAlpha_only or 1=fix_steps, 2=dynamic_search

double usr_angle(int bond, int term13, double *alpha, int numalpha)
{

  double mult = 0.0;  // default for term13 = {10,11,12,13}
  double       phi2 =   -2*alpha[0];
  double       phi1 = M_PI+alpha[0];  // diagonal J' 
  
  if (bond==1) {   // J' bit now...
    // delta term
    // here it's J' parameter model for the 2D Shastry-Sutherand model
    // now the J' part
    if      (term13 == 1) mult=       cos(phi2);    // s^z s^z 
    else if (term13 == 2) mult=  -0.5*sin(phi2);    // s^z s^+ 
    else if (term13 == 3) mult=  -0.5*sin(phi2);    // s^z s^- 
    else if (term13 == 4) mult=   0.5*sin(phi2);    // s^+ s^z
    else if (term13 == 5) mult=   0.5*sin(phi2);    // s^- s^z
    else if (term13 == 6) mult= 0.25*(cos(phi2)+1); // s^+ s^-
    else if (term13 == 7) mult= 0.25*(cos(phi2)+1); // s^- s^+
    else if (term13 == 8) mult= 0.25*(cos(phi2)-1); // s^+ s^+
    else if (term13 == 9) mult= 0.25*(cos(phi2)-1); // s^- s^-
  } else  {   // J bit now... J=1
    if      (term13 == 1) mult=       cos(phi1); // Oct05: -cos 
    else if (term13 == 2) mult=  -0.5*sin(phi1);
    else if (term13 == 3) mult=  -0.5*sin(phi1);
    else if (term13 == 4) mult=   0.5*sin(phi1);
    else if (term13 == 5) mult=   0.5*sin(phi1);
    else if (term13 == 6) mult= 0.25*(cos(phi1)+1);
    else if (term13 == 7) mult= 0.25*(cos(phi1)+1);
    else if (term13 == 8) mult= 0.25*(cos(phi1)-1);
    else if (term13 == 9) mult= 0.25*(cos(phi1)-1);
  }
  return mult;
}
